import cocotb

from typing import Tuple
from enum import IntEnum

import random

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

from cocotbext.axi import AxiLiteMaster, AxiLiteBus

import cocotb.utils

import logging

CLK_PERIOD_200 = 5

ADDR_WIDTH=25
DATA_WIDTH=32

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(dut.i_clk_sys, CLK_PERIOD_200, units="ns").start())
        self.axil = AxiLiteMaster(AxiLiteBus.from_prefix(dut, ""), dut.i_clk_sys, dut.i_rst_sys)

    async def cycle_reset(self):
        self.dut.i_rst_sys.setimmediatevalue(0)
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 0
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)

def decode_wreq(val: int) -> Tuple[int, int]:
    data = (val >> ADDR_WIDTH) & ((1 << DATA_WIDTH)-1)
    addr = val & ((1 << ADDR_WIDTH) - 1)

    return addr, data

@cocotb.test()
async def test_single_write(dut):
    tb = TB(dut)

    tb.dut.wreq_fifo_full.value = 0
    tb.dut.wresp_fifo_empty.value = 1

    await tb.cycle_reset()

    ADDR = 0xaab0
    DATA = 0x11111111

    write = cocotb.start_soon(tb.axil.write_dword(0xaab0, 0x11111111))

    await RisingEdge(tb.dut.wreq_fifo_wr)
    req = tb.dut.wreq_fifo_data.value.integer

    addr, data = decode_wreq(req)

    tb.log.info(f"addr: {addr:x} data: {data:x}")

    assert addr == ADDR
    assert data == DATA

    for _ in range(10):
        await RisingEdge(tb.dut.i_clk_sys)

    tb.log.info("Sending Response")
    tb.dut.wresp_fifo_data.value = 0
    tb.dut.wresp_fifo_empty.value = 0

    await RisingEdge(tb.dut.wresp_fifo_rd)
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.wresp_fifo_data.value = 0
    tb.dut.wresp_fifo_empty.value = 1

    await write

@cocotb.test()
async def test_multiple_write(dut):
    tb = TB(dut)

    tb.dut.wreq_fifo_full.value = 0
    tb.dut.wresp_fifo_empty.value = 1

    await tb.cycle_reset()

    N = 32

    addrs = [random.randint(0, 2**23-1)*4 for _ in range(N)]
    datas = [random.randint(0, 2**32-1) for _ in range(N)]

    writes = [cocotb.start_soon(tb.axil.write_dword(addr, data)) for addr, data in zip(addrs, datas)]

    async def req_resp_thread():
        wr = 0

        while True:
            tb.dut.wresp_fifo_empty.value = 1
            await RisingEdge(tb.dut.wreq_fifo_wr)
            req = tb.dut.wreq_fifo_data.value.integer
            await RisingEdge(tb.dut.i_clk_sys)
            tb.dut.wreq_fifo_full.value = 1

            addr, data = decode_wreq(req)

            tb.log.info(f"addr: {addr:x} data: {data:x}")

            assert addr == addrs[wr]
            assert data == datas[wr]

            wr += 1


            for _ in range(10):
                await RisingEdge(tb.dut.i_clk_sys)

            tb.log.info("Sending Response")
            tb.dut.wresp_fifo_data.value = 0
            tb.dut.wresp_fifo_empty.value = 0

            await RisingEdge(tb.dut.wresp_fifo_rd)
            await RisingEdge(tb.dut.i_clk_sys)
            tb.dut.wresp_fifo_data.value = 0
            tb.dut.wresp_fifo_empty.value = 1
            tb.dut.wreq_fifo_full.value = 0


    cocotb.start_soon(req_resp_thread())
    for write in writes:
        await write

@cocotb.test()
async def test_single_read(dut):
    tb = TB(dut)
    
    tb.dut.rreq_fifo_full.value = 0
    tb.dut.rresp_fifo_empty.value = 1

    await tb.cycle_reset()

    read = cocotb.start_soon(tb.axil.read_dword(0xaab0))

    await RisingEdge(tb.dut.rreq_fifo_wr)
    req = tb.dut.rreq_fifo_data.value.integer

    addr = req

    tb.log.info(f"addr: {addr:x}")

    for _ in range(10):
        await RisingEdge(tb.dut.i_clk_sys)

    tb.log.info("Sending Response")
    data = random.randint(0, 2**32-1)
    tb.dut.rresp_fifo_data.value = data
    tb.dut.rresp_fifo_empty.value = 0

    await RisingEdge(tb.dut.rresp_fifo_rd)
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.rresp_fifo_data.value = 0
    tb.dut.rresp_fifo_empty.value = 1

    rd_data = await read

    assert rd_data == data

@cocotb.test()
async def test_multiple_read(dut):
    tb = TB(dut)
    
    tb.dut.rreq_fifo_full.value = 0
    tb.dut.rresp_fifo_empty.value = 1

    await tb.cycle_reset()

    N = 32
    addrs = [random.randint(0, 2**23-1)*4 for _ in range(N)]
    datas = [random.randint(0, 2**32-1) for _ in range(N)]

    reads = [cocotb.start_soon(tb.axil.read_dword(addr)) for addr in addrs]

    async def req_resp_thread():
        rd = 0

        while True:
            tb.dut.rresp_fifo_empty.value = 1
            await RisingEdge(tb.dut.rreq_fifo_wr)
            req = tb.dut.rreq_fifo_data.value.integer
            await RisingEdge(tb.dut.i_clk_sys)
            tb.dut.rreq_fifo_full.value = 1

            addr = req

            tb.log.info(f"addr: {addr:x}")

            assert addr == addrs[rd]

            for _ in range(10):
                await RisingEdge(tb.dut.i_clk_sys)

            tb.log.info("Sending Response")
            tb.dut.rresp_fifo_data.value = datas[rd]
            tb.dut.rresp_fifo_empty.value = 0

            await RisingEdge(tb.dut.rresp_fifo_rd)
            await RisingEdge(tb.dut.i_clk_sys)
            tb.dut.rresp_fifo_data.value = 0
            tb.dut.rresp_fifo_empty.value = 1
            tb.dut.rreq_fifo_full.value = 0

            rd += 1

    cocotb.start_soon(req_resp_thread())
    for read, data in zip(reads, datas):
        rd_data = await read
        assert rd_data == data
