import cocotb

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import cocotb.utils

import logging

import math

from typing import Dict


CLK_PERIOD_200 = 5

TPOWERUP_NS = 200000
TREF_MS = 64

TRP_NS = 15
TMRD_NS = 10

TRP_CLK = math.ceil(TRP_NS/CLK_PERIOD_200)
TMRD_CLK = math.ceil(TMRD_NS/CLK_PERIOD_200)

ADDR_WIDTH=25
DATA_WIDTH=32

cmd_names: Dict[int, str] = {
    8: "inhibit",
    7: "nop",
    6: "burst_term",
    5: "read",
    4: "write",
    3: "activate",
    2: "precharge",
    1: "auto_refresh",
    0: "mode_set"
}

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(dut.i_clk_sys, CLK_PERIOD_200, units="ns").start())

    async def cycle_reset(self):
        self.dut.rreq_fifo_empty.value = 1
        self.dut.wreq_fifo_empty.value = 1
        self.dut.i_cmd_ready.value = 0

        self.dut.i_rst_sys.setimmediatevalue(0)
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 0
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)

    async def wait_for_cmd(self) -> int:
        self.dut.i_cmd_ready.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        while not self.dut.o_cmd_valid.value:
            await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_cmd_ready.value = 0
        return self.dut.o_cmd.value.integer

    async def ref_req_model(self):
        self.dut.i_ref_req.value = 0
        while True:
            for _ in range(TREF_CLOCKS-1):
                await RisingEdge(self.dut.i_clk_sys)
            self.dut.i_ref_req.value = 1
            await RisingEdge(self.dut.i_clk_sys)
            self.dut.i_ref_req.value = 0

    async def send_write_req(self, waddr, wdata):
        self.dut.wreq_fifo_data.value = (wdata << ADDR_WIDTH) | waddr
        self.dut.wreq_fifo_empty.value = 0
        while True:
            await RisingEdge(self.dut.i_clk_sys)
            if self.dut.wreq_fifo_rd.value:
                self.dut.wreq_fifo_empty.value = 1
                break

    async def send_read_req(self, waddr):
        self.dut.rreq_fifo_data.value = waddr
        self.dut.rreq_fifo_empty.value = 0
        while True:
            await RisingEdge(self.dut.i_clk_sys)
            if self.dut.rreq_fifo_rd.value:
                self.dut.rreq_fifo_empty.value = 1
                break

    async def initialize(self):
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_init_req.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_init_req.value = 0

        self.dut.i_cmd_ready.value = 1
        while True:
            await RisingEdge(self.dut.i_clk_sys)
            if self.dut.init_state.value == 5:
                break
        self.dut.i_cmd_ready.value = 0

TREF_NS = TREF_MS * 1000000
TREF_MS = 64
REFRESH_COUNT = 8192
TPOWERUP_CLOCKS = math.ceil(TPOWERUP_NS/CLK_PERIOD_200)
TREF_CLOCKS = math.floor((TREF_NS/REFRESH_COUNT)/CLK_PERIOD_200)

@cocotb.test()
async def test_pwrup(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    async def pwrup_phy_model():
        start_time = cocotb.utils.get_sim_time()//1000
        await RisingEdge(tb.dut.o_cmd_valid)
        tpwrup = cocotb.utils.get_sim_time()//1000 - start_time

        tb.log.info(f"tpwrup: {tpwrup}ns")
        assert tpwrup >= TPOWERUP_NS

        async def wait_for_cmd() -> int:
            tb.dut.i_cmd_ready.value = 1
            await RisingEdge(tb.dut.i_clk_sys)
            while not tb.dut.o_cmd_valid.value:
                await RisingEdge(tb.dut.i_clk_sys)
            tb.dut.i_cmd_ready.value = 0
            return tb.dut.o_cmd.value.integer

        cmd = await wait_for_cmd()
        tb.log.info(f"cmd: {cmd_names[cmd]}")
        assert cmd_names[cmd] == "precharge"

        for _ in range(TRP_CLK):
            await RisingEdge(tb.dut.i_clk_sys)

        cmd = await wait_for_cmd()
        tb.log.info(f"cmd: {cmd_names[cmd]}")
        assert cmd_names[cmd] == "mode_set"

        for _ in range(TMRD_CLK):
            await RisingEdge(tb.dut.i_clk_sys)


        while True:
            cmd = await wait_for_cmd()
            tb.log.info(f"cmd: {cmd_names[cmd]}")

            if cmd_names[cmd] != "auto_refresh":
                break



    cocotb.start_soon(tb.ref_req_model())
    cocotb.start_soon(pwrup_phy_model())

    await Timer(200, "us")
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_init_req.value = 1
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_init_req.value = 0


    while True:
        await RisingEdge(tb.dut.i_clk_sys)
        if tb.dut.init_state.value == 5:
            break

@cocotb.test()
async def test_ref_req(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_init_req.value = 1
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_init_req.value = 0

    cocotb.start_soon(tb.ref_req_model())

    ref_reqs = 0

    N = 16

    t_start = cocotb.utils.get_sim_time()//1000
    while True:
        cmd = await tb.wait_for_cmd()

        if cmd_names[cmd] == "auto_refresh":
            ref_reqs += 1

            tb.log.info(f"req {ref_reqs}")

        if ref_reqs >= N:
            break

    t_elapsed = cocotb.utils.get_sim_time()//1000 - t_start

    interval = t_elapsed / ref_reqs

    tb.log.info(f"interval: {interval}")
    extrapolated = interval*8192
    tb.log.info(f"extrapolated: {extrapolated}")

    assert extrapolated <= 64000000

    tb.log.info(f"elapsed: {t_elapsed}")

@cocotb.test()
async def test_write(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    cocotb.start_soon(tb.ref_req_model())
    await tb.initialize()

    cocotb.start_soon(tb.send_write_req(0, 0))

    while True:
        cmd = await tb.wait_for_cmd()
        tb.log.info(f"cmd: {cmd_names[cmd]}")

        if cmd_names[cmd] == "write":
            break

    await Timer(1000, "ns")


@cocotb.test()
async def test_read(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    cocotb.start_soon(tb.ref_req_model())
    await tb.initialize()

    cocotb.start_soon(tb.send_read_req(0))

    while True:
        cmd = await tb.wait_for_cmd()
        tb.log.info(f"cmd: {cmd_names[cmd]}")

        if cmd_names[cmd] == "read":
            break

    await Timer(1000, "ns")

@cocotb.test()
async def test_ref_priority(dut):
    """Make sure that refresh requests have priority"""

    tb = TB(dut)

    await tb.cycle_reset()

    tb.dut.i_ref_req.value = 1

    await tb.initialize()

    tb.dut.i_ref_req.value = 0

    tb.dut.i_cmd_ready.value = 1
    tb.dut.wreq_fifo_empty.value = 0

    await Timer(100, "ns")
    await RisingEdge(tb.dut.i_clk_sys)
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_ref_req.value = 1
    await RisingEdge(tb.dut.i_clk_sys)
    tb.dut.i_ref_req.value = 0

    await Timer(1000, "ns")