import cocotb

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import cocotb.utils

import logging


CLK_PERIOD_200 = 5

TPOWERUP_NS = 200000
TREF_MS = 64

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(dut.i_clk_sys, CLK_PERIOD_200, units="ns").start())


    async def cycle_reset(self):
        self.dut.i_rst_sys.setimmediatevalue(0)
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 0
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)

@cocotb.test()
async def test_pwrup(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    RESET_TIME = CLK_PERIOD_200*3

    await RisingEdge(tb.dut.o_init_req)
    elapsed = cocotb.utils.get_sim_time()//1000 - RESET_TIME

    tb.log.info(f"Powerup Time: {elapsed}")
    assert elapsed == TPOWERUP_NS

    await Timer(1000, "ns")

@cocotb.test()
async def test_ref_req(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    await RisingEdge(tb.dut.o_ref_req)
    start_time = cocotb.utils.get_sim_time()//1000

    for _ in range(16):
        await RisingEdge(tb.dut.o_ref_req)
        elapsed = cocotb.utils.get_sim_time()//1000 - start_time
        start_time = cocotb.utils.get_sim_time()//1000

        tb.log.info(f"Elapsed: {elapsed}")

        extrapolated = elapsed*8192

        tb.log.info(f"Extrapolated: {extrapolated}")

        assert extrapolated / 1000000 <= TREF_MS
