import cocotb

from cocotb.clock import Clock
from cocotb.triggers import Timer, RisingEdge, FallingEdge

import cocotb.utils

import logging

CLK_PERIOD_200 = 5

CAS_LATENCY=3

class TB:
    def __init__(self, dut):
        self.dut = dut

        self.log = logging.getLogger("cocotb.tb")
        self.log.setLevel(logging.INFO)

        cocotb.start_soon(Clock(dut.i_clk_sys, CLK_PERIOD_200, units="ns").start())


    async def cycle_reset(self):
        self.dut.i_rst_sys.setimmediatevalue(0)
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 1
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)
        self.dut.i_rst_sys.value = 0
        await RisingEdge(self.dut.i_clk_sys)
        await RisingEdge(self.dut.i_clk_sys)

    async def send_cmd(self, cmd, ba=0, addr=0, data=0):
        self.dut.i_cmd.value = cmd
        self.dut.i_cmd_valid.value = 1
        self.dut.i_ba.value = ba
        self.dut.i_addr.value = addr
        self.dut.i_data.value = data
        while True:
            await RisingEdge(self.dut.i_clk_sys)
            if self.dut.o_cmd_ready.value:
                break
        self.dut.i_cmd.value = 0
        self.dut.i_cmd_valid.value = 0
        self.dut.i_ba.value = 0
        self.dut.i_addr.value = 0
        self.dut.i_data.value = 0

    async def send_read(self, row, col, sim_data):
        await self.send_cmd(3, addr=row)
        await self.send_cmd(5, addr=(1 << 10) | col)

        for _ in range(CAS_LATENCY):
            await FallingEdge(self.dut.i_clk_sys)

        self.dut.i_sdram_dq.value = sim_data >> 16
        await FallingEdge(self.dut.i_clk_sys)
        self.dut.i_sdram_dq.value = sim_data & ((1 << 16) -1)
        await RisingEdge(self.dut.i_clk_sys)
        assert self.dut.o_data_valid.value == 1
        data = self.dut.o_data.value.integer
        await FallingEdge(self.dut.i_clk_sys)
        self.dut.i_sdram_dq.value = 0
        return data

@cocotb.test()
async def test_pwrup(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    await tb.send_cmd(2, addr=1 << 10)
    await tb.send_cmd(0, addr=((3 << 4) | (1 << 0)))
    await tb.send_cmd(1)
    await tb.send_cmd(1)

    await Timer(1000, "ns")

@cocotb.test()
async def test_write(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    await tb.send_cmd(3, addr=0xabc)
    await tb.send_cmd(4, addr=(1 << 10) | 1, data=0xabcd1234)

    await tb.send_cmd(3, addr=0x123)
    await tb.send_cmd(4, addr=(1 << 10) | 2, data=0xabcd1234)

    await tb.send_cmd(3, addr=0xabc)
    await tb.send_cmd(4, addr=(1 << 10) | 3, data=0xabcd1234)

    await Timer(1000, "ns")

@cocotb.test()
async def test_read(dut):
    tb = TB(dut)

    await tb.cycle_reset()

    data = await tb.send_read(row=0xabc, col=0xaa, sim_data=0xaaaa5555)
    tb.log.info(f"Data: {data:x}")

    await Timer(1000, "ns")
