module sdram_axil_interface #(
    parameter AXIL_DATA_WIDTH = 32,
    parameter AXIL_ADDR_WIDTH = 25, // byte address,
    localparam AXIL_STRB_WIDTH=(AXIL_DATA_WIDTH+7)/8,

    parameter WREQ_FIFO_WIDTH=AXIL_DATA_WIDTH+AXIL_ADDR_WIDTH,
    parameter RREQ_FIFO_WIDTH=AXIL_ADDR_WIDTH,
    parameter WRESP_FIFO_WIDTH=2,
    parameter RRESP_FIFO_WIDTH=AXIL_DATA_WIDTH+2
) (
    input  logic                        i_clk_sys,
    input  logic                        i_rst_sys,

    output logic                        awready,
    input  logic                        awvalid,
    input  logic [AXIL_ADDR_WIDTH-1:0]  awaddr,
    input  logic [2:0]                  awprot,
    output logic                        wready,
    input  logic                        wvalid,
    input  logic [AXIL_DATA_WIDTH-1:0]  wdata,
    input  logic [AXIL_STRB_WIDTH-1:0]  wstrb,
    input  logic                        bready,
    output logic                        bvalid,
    output logic [1:0]                  bresp,
    output logic                        arready,
    input  logic                        arvalid,
    input  logic [AXIL_ADDR_WIDTH-1:0]  araddr,
    input  logic [2:0]                  arprot,
    input  logic                        rready,
    output logic                        rvalid,
    output logic [AXIL_DATA_WIDTH-1:0]  rdata,
    output logic [1:0]                  rresp,

    output logic [WREQ_FIFO_WIDTH-1:0]  wreq_fifo_data,
    output logic                        wreq_fifo_wr,
    input  logic                        wreq_fifo_full,

    output logic [RREQ_FIFO_WIDTH-1:0]  rreq_fifo_data,
    output logic                        rreq_fifo_wr,
    input  logic                        rreq_fifo_full,

    input  logic [WRESP_FIFO_WIDTH-1:0] wresp_fifo_data,
    output logic                        wresp_fifo_rd,
    input  logic                        wresp_fifo_empty,

    input  logic [RRESP_FIFO_WIDTH-1:0] rresp_fifo_data,
    output logic                        rresp_fifo_rd,
    input  logic                        rresp_fifo_empty
);



`ifdef SIM
initial begin
    $display("WREQ_FIFO_WIDTH: %d", WREQ_FIFO_WIDTH);
    $display("WRESP_FIFO_WIDTH: %d", WRESP_FIFO_WIDTH);
    $display("RREQ_FIFO_WIDTH: %d", RREQ_FIFO_WIDTH);
    $display("RRESP_FIFO_WIDTH: %d", RRESP_FIFO_WIDTH);
end
`endif

// Write section

logic [AXIL_ADDR_WIDTH-1:0] waddr_reg, waddr_reg_next;
logic saw_awvalid, saw_awvalid_next;
logic [AXIL_DATA_WIDTH-1:0] wdata_reg, wdata_reg_next;
logic saw_wvalid, saw_wvalid_next;

always_ff @(posedge i_clk_sys) begin : WRITE_FF
    if (i_rst_sys) begin
        waddr_reg   <= '0;
        wdata_reg   <= '0;
        saw_awvalid <= '0;
        saw_wvalid  <= '0;
    end else begin
        waddr_reg   <= waddr_reg_next;
        wdata_reg   <= wdata_reg_next;
        saw_awvalid <= saw_awvalid_next;
        saw_wvalid  <= saw_wvalid_next;
    end
end

always_comb begin : WRITE_COMB
    awready = !saw_awvalid;
    wready = !saw_wvalid;

    saw_awvalid_next = saw_awvalid;
    saw_wvalid_next = saw_wvalid;

    waddr_reg_next = waddr_reg;
    wdata_reg_next = wdata_reg;

    wreq_fifo_wr = '0;
    wreq_fifo_data = '0;
    wresp_fifo_rd = '0;

    bresp = '0;
    bvalid = '0;

    if (awvalid & awready) begin
        saw_awvalid_next = '1;
        waddr_reg_next = awaddr;
    end

    if (wvalid & wready) begin
        saw_wvalid_next = '1;
        wdata_reg_next = wdata;
    end

    // we can reclaim a cycle here
    if (saw_awvalid & saw_wvalid) begin
        if (!wreq_fifo_full) begin
            wreq_fifo_data = {wdata_reg, waddr_reg};
            wreq_fifo_wr = '1;

            saw_awvalid_next = '0;
            saw_wvalid_next = '0;
        end
    end


    if (!wresp_fifo_empty) begin
        bresp = wresp_fifo_data;
        bvalid = '1;

        if (bready) begin
            wresp_fifo_rd = '1;
        end
    end
end


// Read section

logic saw_arvalid, saw_arvalid_next;
logic [AXIL_ADDR_WIDTH-1:0] raddr_reg, raddr_reg_next;

always_ff @(posedge i_clk_sys) begin : READ_FF
    if (i_rst_sys) begin
        raddr_reg   <= '0;
        saw_arvalid <= '0;
    end else begin
        raddr_reg   <= raddr_reg_next;
        saw_arvalid <= saw_arvalid_next;
    end
end

always_comb begin : READ_COMB
    arready = !saw_arvalid;

    saw_arvalid_next = saw_arvalid;

    raddr_reg_next = raddr_reg;

    rreq_fifo_data = '0;
    rreq_fifo_wr = '0;

    rresp_fifo_rd = '0;

    rresp = '0;
    rdata = '0;
    rvalid = '0;

    if (arvalid & arready) begin
        saw_arvalid_next = '1;
        raddr_reg_next = araddr;
    end

    // we can reclaim a cycle here
    if (saw_arvalid) begin
        if (!rreq_fifo_full) begin
            rreq_fifo_data = {raddr_reg};
            rreq_fifo_wr = '1;

            saw_arvalid_next = '0;
        end
    end

    if (!rresp_fifo_empty) begin
        rresp = rresp_fifo_data[RRESP_FIFO_WIDTH-1:RRESP_FIFO_WIDTH-2];
        rdata = rresp_fifo_data[AXIL_DATA_WIDTH-1:0];
        rvalid = '1;

        if (rready) begin
            rresp_fifo_rd = '1;
        end
    end
end


endmodule