module sdram_command_generator #(
    parameter AXIL_DATA_WIDTH = 32,
    parameter AXIL_ADDR_WIDTH = 25,

    parameter WREQ_FIFO_WIDTH=AXIL_DATA_WIDTH+AXIL_ADDR_WIDTH,
    parameter RREQ_FIFO_WIDTH=AXIL_ADDR_WIDTH,

    localparam CMD_WIDTH=4,
    parameter BA_WIDTH=2,
    parameter ADDR_WIDTH=13,
    parameter COL_WIDTH=9,
    parameter CAS_LATENCY=3,
    parameter BURST_LENGTH=2,
    parameter INIT_REFRESH_CYCLES=3
) (
    input  logic                        i_clk_sys,
    input  logic                        i_rst_sys,

    input  logic                        i_init_req,
    input  logic                        i_ref_req,

    input  logic [WREQ_FIFO_WIDTH-1:0]  wreq_fifo_data,
    output logic                        wreq_fifo_rd,
    input  logic                        wreq_fifo_empty,

    input  logic [RREQ_FIFO_WIDTH-1:0]  rreq_fifo_data,
    output logic                        rreq_fifo_rd,
    input  logic                        rreq_fifo_empty,

    output logic [CMD_WIDTH-1:0]        o_cmd,
    output logic [BA_WIDTH-1:0]         o_ba,
    output logic [ADDR_WIDTH-1:0]       o_addr,
    output logic [AXIL_DATA_WIDTH-1:0]  o_data,
    output logic                        o_cmd_valid,
    input  logic                        i_cmd_ready
);

enum logic [2:0] {IDLE, PRECHARGE_ALL, SET_MODE, REFRESH_1, REFRESH_2, DONE} init_state, init_state_next;


logic [15:0] init_ref_req_counter, init_ref_req_counter_next;
logic ref_req_pending, ref_req_pending_next;

enum logic [1:0] {NONE, READ, WRITE} op, op_reg, op_next, last_op, last_op_next;

enum logic {ROW, COL} access_state, access_state_next;

logic [BA_WIDTH-1:0] ba_reg, ba_reg_next;
logic [COL_WIDTH-1:0] col_reg, col_reg_next;
logic [AXIL_DATA_WIDTH-1:0] data_reg, data_reg_next;

always_ff @(posedge i_clk_sys) begin
    if (i_rst_sys) begin
        init_state <= IDLE;
        init_ref_req_counter <= '0;
        last_op <= READ;
        op_reg <= NONE;
        access_state <= ROW;
        ref_req_pending <= '0;
        data_reg <= '0;
    end else begin
        init_state <= init_state_next;
        init_ref_req_counter <= init_ref_req_counter_next;
        last_op <= last_op_next;
        op_reg <= op_next;
        access_state <= access_state_next;
        ref_req_pending <= ref_req_pending_next;
        data_reg <= data_reg_next;
    end
end

// verilator lint_off UNOPTFLAT
always_comb begin
    init_state_next = init_state;
    init_ref_req_counter_next = init_ref_req_counter;

    o_cmd = 4'b1000;
    o_cmd_valid = '0;
    o_ba = '0;
    o_addr = '0;

    ref_req_pending_next = ref_req_pending;

    wreq_fifo_rd = '0;
    rreq_fifo_rd = '0;

    op = NONE;
    op_next = NONE;

    access_state_next = access_state;

    case (init_state)
        IDLE: begin
            if (i_init_req) begin
                init_state_next = PRECHARGE_ALL;
            end
        end

        PRECHARGE_ALL: begin
            o_cmd = 4'h2;
            o_cmd_valid = '1;
            o_addr[10] = 1;

            if (i_cmd_ready) begin
                init_state_next = SET_MODE;
            end
        end

        SET_MODE: begin
            o_cmd = '0;
            o_cmd_valid = '1;
            o_ba = '0;
            o_addr = {{ADDR_WIDTH-10{1'b0}}, 1'b0, 2'b0, 3'(CAS_LATENCY), 1'b0, 3'($clog2(BURST_LENGTH))};

            if (i_cmd_ready) begin
                init_state_next = REFRESH_1;
            end
        end

        REFRESH_1: begin
            if (i_ref_req) begin
                init_state_next = REFRESH_2;
            end
        end

        REFRESH_2: begin
            o_cmd = 4'h1;
            o_cmd_valid = '1;

            if (i_cmd_ready) begin
                init_ref_req_counter_next = init_ref_req_counter + 1;

                if (init_ref_req_counter_next == INIT_REFRESH_CYCLES) begin
                    init_state_next = DONE;
                end else begin
                    init_state_next = REFRESH_1;
                end
            end
        end

        DONE: begin
            init_state_next = DONE;
        end

        default: begin
            `ifdef SIM
                $error("Invalid SDRAM Init State");
            `endif
            // just go to DONE and do nothing.
            init_state_next = DONE;
        end

    endcase

    if (init_state == DONE) begin
        if (access_state == ROW) begin
            if (i_ref_req || ref_req_pending) begin
                o_cmd = 4'h1;
                o_cmd_valid = '1;

                if (i_cmd_ready) begin
                    ref_req_pending_next = '0;
                end else begin
                    ref_req_pending_next = '1;
                end
            end else begin
                // arbitrate requests
                if (last_op == READ) begin
                    if (!wreq_fifo_empty) begin
                        op = WRITE;
                    end else if (!rreq_fifo_empty) begin
                        op = READ;
                    end
                end else begin
                    if (!rreq_fifo_empty) begin
                        op = READ;
                    end else if (!wreq_fifo_empty) begin
                        op = WRITE;
                    end
                end
            end

            case (op)
                NONE: begin
                    // do nothing.
                end 

                READ: begin
                    op_next = READ;
                    ba_reg_next = rreq_fifo_data[BA_WIDTH+ADDR_WIDTH+COL_WIDTH-1:ADDR_WIDTH+COL_WIDTH];
                    col_reg_next = rreq_fifo_data[COL_WIDTH-1:0];

                    o_cmd = 4'h3;
                    o_cmd_valid = '1;
                    o_ba = rreq_fifo_data[BA_WIDTH+ADDR_WIDTH+COL_WIDTH-1:ADDR_WIDTH+COL_WIDTH];
                    o_addr = rreq_fifo_data[ADDR_WIDTH+COL_WIDTH-1:COL_WIDTH];

                    if (i_cmd_ready) begin
                        rreq_fifo_rd = '1;
                        access_state_next = COL;
                    end
                end

                WRITE: begin
                    op_next = WRITE;
                    ba_reg_next = wreq_fifo_data[BA_WIDTH+ADDR_WIDTH+COL_WIDTH-1:ADDR_WIDTH+COL_WIDTH];
                    col_reg_next = wreq_fifo_data[COL_WIDTH-1:0];
                    data_reg_next = wreq_fifo_data[AXIL_DATA_WIDTH+AXIL_ADDR_WIDTH-1:AXIL_ADDR_WIDTH];

                    o_cmd = 4'h3;
                    o_cmd_valid = '1;
                    o_ba = wreq_fifo_data[BA_WIDTH+ADDR_WIDTH+COL_WIDTH-1:ADDR_WIDTH+COL_WIDTH];
                    o_addr = wreq_fifo_data[ADDR_WIDTH+COL_WIDTH-1:COL_WIDTH];

                    if (i_cmd_ready) begin
                        wreq_fifo_rd = '1;
                        access_state_next = COL;
                    end
                end

                default: begin
                    `ifdef SIM
                    $error("Invalid SDRAM OP");
                    `endif
                end
            endcase 
        end else if (access_state == COL) begin
            if (i_ref_req) begin
                ref_req_pending_next = '1;
            end

            if (op_reg == WRITE) begin
                o_cmd = 4'h4;
                o_data = data_reg;
            end else if (op_reg == READ) begin
                o_cmd = 4'h5;
            end
            o_addr[COL_WIDTH-1:0] = col_reg;
            o_addr[10] = 1;
            o_cmd_valid = '1;

            if (i_cmd_ready) begin
                access_state_next = ROW;
                op_next = NONE;
            end else begin
                op_next = op_reg;
            end
        end
    end
end

endmodule