module sdram_controller #(
    parameter AXIL_DATA_WIDTH = 32,
    parameter AXIL_ADDR_WIDTH = 25, // byte address,
    localparam AXIL_STRB_WIDTH=(AXIL_DATA_WIDTH+7)/8,

    parameter CAS_LATENCY=3,
    parameter BURST_LENGTH=2,
    parameter TRAS_NS=40,
    parameter TRC_NS=55,
    parameter TRCD_NS=15,
    parameter TRFC_NS=55,
    parameter TRP_NS=15,
    parameter TWR_NS=10,
    parameter TMRD_NS=10,

    parameter TPOWERUP_NS = 200000,
    parameter TREF_MS = 64,
    parameter REFRESH_COUNT = 8192,

    parameter DQ_WIDTH=16,

    parameter BA_WIDTH=2,
    parameter ADDR_WIDTH=13,
    parameter COL_WIDTH=9
) (
    input  logic                        i_clk_sys,
    input  logic                        i_rst_sys,

    output logic                        awready,
    input  logic                        awvalid,
    input  logic [AXIL_ADDR_WIDTH-1:0]  awaddr,
    input  logic [2:0]                  awprot,
    output logic                        wready,
    input  logic                        wvalid,
    input  logic [AXIL_DATA_WIDTH-1:0]  wdata,
    input  logic [AXIL_STRB_WIDTH-1:0]  wstrb,
    input  logic                        bready,
    output logic                        bvalid,
    output logic [1:0]                  bresp,
    output logic                        arready,
    input  logic                        arvalid,
    input  logic [AXIL_ADDR_WIDTH-1:0]  araddr,
    input  logic [2:0]                  arprot,
    input  logic                        rready,
    output logic                        rvalid,
    output logic [AXIL_DATA_WIDTH-1:0]  rdata,
    output logic [1:0]                  rresp,

    output logic [DQ_WIDTH-1:0]         o_sdram_dq,
    output logic [DQ_WIDTH-1:0]         o_sdram_dq_oe,
    input  logic [DQ_WIDTH-1:0]         i_sdram_dq,
    output logic [ADDR_WIDTH-1:0]       o_sdram_addr,
    output logic [BA_WIDTH-1:0]         o_sdram_ba,
    output logic                        o_sdram_cke,
    output logic                        o_sdram_cs_n,
    output logic                        o_sdram_ras_n,
    output logic                        o_sdram_cas_n,
    output logic                        o_sdram_we_n
);

localparam WREQ_FIFO_WIDTH=AXIL_DATA_WIDTH+AXIL_ADDR_WIDTH;
localparam RREQ_FIFO_WIDTH=AXIL_ADDR_WIDTH;
localparam WRESP_FIFO_WIDTH=2;
localparam RRESP_FIFO_WIDTH=AXIL_DATA_WIDTH+2;

logic [WREQ_FIFO_WIDTH-1:0]     wreq_fifo_i_data;
logic [WREQ_FIFO_WIDTH-1:0]     wreq_fifo_o_data;
logic                           wreq_fifo_wr;
logic                           wreq_fifo_rd;
logic                           wreq_fifo_full;
logic                           wreq_fifo_empty;

logic [RREQ_FIFO_WIDTH-1:0]     rreq_fifo_i_data;
logic [RREQ_FIFO_WIDTH-1:0]     rreq_fifo_o_data;
logic                           rreq_fifo_wr;
logic                           rreq_fifo_rd;
logic                           rreq_fifo_full;
logic                           rreq_fifo_empty;

logic [WRESP_FIFO_WIDTH-1:0]    wresp_fifo_i_data;
logic [WRESP_FIFO_WIDTH-1:0]    wresp_fifo_o_data;
logic                           wresp_fifo_wr;
logic                           wresp_fifo_rd;
logic                           wresp_fifo_full;
logic                           wresp_fifo_empty;

logic [RRESP_FIFO_WIDTH-1:0]    rresp_fifo_i_data;
logic [RRESP_FIFO_WIDTH-1:0]    rresp_fifo_o_data;
logic                           rresp_fifo_wr;
logic                           rresp_fifo_rd;
logic                           rresp_fifo_full;
logic                           rresp_fifo_empty;

logic                           ref_req;
logic                           init_req;

logic [3:0]                     phy_cmd;
logic [BA_WIDTH-1:0]            phy_ba;
logic [ADDR_WIDTH-1:0]          phy_addr;
logic [AXIL_DATA_WIDTH-1:0]     phy_data;
logic                           phy_cmd_valid;
logic                           phy_cmd_ready;

sdram_axil_interface #(
    .AXIL_DATA_WIDTH    (AXIL_DATA_WIDTH),
    .AXIL_ADDR_WIDTH    (AXIL_ADDR_WIDTH),

    .WREQ_FIFO_WIDTH    (WREQ_FIFO_WIDTH),
    .RREQ_FIFO_WIDTH    (RREQ_FIFO_WIDTH),
    .WRESP_FIFO_WIDTH   (WRESP_FIFO_WIDTH),
    .RRESP_FIFO_WIDTH   (RRESP_FIFO_WIDTH)
) u_sdram_axil_interface (
    .i_clk_sys          (i_clk_sys),
    .i_rst_sys          (i_rst_sys),

    .awready            (awready),
    .awvalid            (awvalid),
    .awaddr             (awaddr),
    .awprot             (awprot),
    .wready             (wready),
    .wvalid             (wvalid),
    .wdata              (wdata),
    .wstrb              (wstrb),
    .bready             (bready),
    .bvalid             (bvalid),
    .bresp              (bresp),
    .arready            (arready),
    .arvalid            (arvalid),
    .araddr             (araddr),
    .arprot             (arprot),
    .rready             (rready),
    .rvalid             (rvalid),
    .rdata              (rdata),
    .rresp              (rresp),

    .wreq_fifo_data     (wreq_fifo_i_data),
    .wreq_fifo_wr       (wreq_fifo_wr),
    .wreq_fifo_full     (wreq_fifo_full),

    .rreq_fifo_data     (rreq_fifo_i_data),
    .rreq_fifo_wr       (rreq_fifo_wr),
    .rreq_fifo_full     (rreq_fifo_full),

    .wresp_fifo_data    (wresp_fifo_o_data),
    .wresp_fifo_rd      (wresp_fifo_rd),
    .wresp_fifo_empty   (wresp_fifo_empty),

    .rresp_fifo_data    (rresp_fifo_o_data),
    .rresp_fifo_rd      (rresp_fifo_rd),
    .rresp_fifo_empty   (rresp_fifo_empty)
);


fifo_fwft #(
    .DATA_WIDTH(WREQ_FIFO_WIDTH),
    .DEPTH_WIDTH(4)
) u_wreq_fifo (
    .clk                (i_clk_sys),
    .rst                (i_rst_sys),
    .din                (wreq_fifo_i_data),
    .wr_en              (wreq_fifo_wr),
    .full               (wreq_fifo_full),
    .dout               (wreq_fifo_o_data),
    .rd_en              (wreq_fifo_rd),
    .empty              (wreq_fifo_empty)
);

fifo_fwft #(
    .DATA_WIDTH(RREQ_FIFO_WIDTH),
    .DEPTH_WIDTH(4)
) u_rreq_fifo (
    .clk                (i_clk_sys),
    .rst                (i_rst_sys),
    .din                (rreq_fifo_i_data),
    .wr_en              (rreq_fifo_wr),
    .full               (rreq_fifo_full),
    .dout               (rreq_fifo_o_data),
    .rd_en              (rreq_fifo_rd),
    .empty              (rreq_fifo_empty)
);


fifo_fwft #(
    .DATA_WIDTH(WRESP_FIFO_WIDTH),
    .DEPTH_WIDTH(4)
) u_wresp_fifo (
    .clk                (i_clk_sys),
    .rst                (i_rst_sys),
    .din                (wresp_fifo_i_data),
    .wr_en              (wresp_fifo_wr),
    .full               (wresp_fifo_full),
    .dout               (wresp_fifo_o_data),
    .rd_en              (wresp_fifo_rd),
    .empty              (wresp_fifo_empty)
);


fifo_fwft #(
    .DATA_WIDTH(RRESP_FIFO_WIDTH),
    .DEPTH_WIDTH(4)
) u_rresp_fifo (
    .clk                (i_clk_sys),
    .rst                (i_rst_sys),
    .din                (rresp_fifo_i_data),
    .wr_en              (rresp_fifo_wr),
    .full               (rresp_fifo_full),
    .dout               (rresp_fifo_o_data),
    .rd_en              (rresp_fifo_rd),
    .empty              (rresp_fifo_empty)
);

sdram_refresh_generator #(
    .TPOWERUP_NS        (200000),
    .TREF_MS            (64),
    .REFRESH_COUNT      (8192)
) u_refresh_generator (
    .i_clk_sys          (i_clk_sys),
    .i_rst_sys          (i_rst_sys),

    .o_ref_req          (ref_req),
    .o_init_req         (init_req)
);

sdram_command_generator #(
    .AXIL_DATA_WIDTH    (AXIL_DATA_WIDTH),
    .AXIL_ADDR_WIDTH    (AXIL_ADDR_WIDTH),

    .WREQ_FIFO_WIDTH    (WREQ_FIFO_WIDTH),
    .RREQ_FIFO_WIDTH    (RREQ_FIFO_WIDTH),

    .BA_WIDTH           (BA_WIDTH),
    .ADDR_WIDTH         (ADDR_WIDTH),
    .COL_WIDTH          (COL_WIDTH),
    .CAS_LATENCY        (CAS_LATENCY),
    .BURST_LENGTH       (BURST_LENGTH),
    .INIT_REFRESH_CYCLES(3)
) u_sdram_command_generator (
    .i_clk_sys          (i_clk_sys),
    .i_rst_sys          (i_rst_sys),

    .i_init_req         (init_req),
    .i_ref_req          (ref_req),

    .wreq_fifo_data     (wreq_fifo_o_data),
    .wreq_fifo_rd       (wreq_fifo_rd),
    .wreq_fifo_empty    (wreq_fifo_empty),

    .rreq_fifo_data     (rreq_fifo_o_data),
    .rreq_fifo_rd       (rreq_fifo_rd),
    .rreq_fifo_empty    (rreq_fifo_empty),

    .o_cmd              (phy_cmd),
    .o_ba               (phy_ba),
    .o_addr             (phy_addr),
    .o_data             (phy_data),
    .o_cmd_valid        (phy_cmd_valid),
    .i_cmd_ready        (phy_cmd_ready)
);

logic phy_data_valid;
logic [2:0] phy_resp;
logic [AXIL_DATA_WIDTH-1:0] phy_o_data;

assign wresp_fifo_wr = phy_data_valid & phy_resp[2];
assign rresp_fifo_wr = phy_data_valid & ~phy_resp[2];

assign wresp_fifo_i_data = phy_resp[1:0];
assign rresp_fifo_i_data = {phy_resp[1:0], phy_o_data};

sdram_phy #(
    .CAS_LATENCY        (CAS_LATENCY),
    .BURST_LENGTH       (BURST_LENGTH),
    .TRAS_NS            (TRAS_NS),
    .TRC_NS             (TRC_NS),
    .TRCD_NS            (TRCD_NS),
    .TRFC_NS            (TRFC_NS),
    .TRP_NS             (TRP_NS),
    .TWR_NS             (TWR_NS),
    .TMRD_NS            (TMRD_NS),

    .DQ_WIDTH           (DQ_WIDTH),

    .BA_WIDTH           (BA_WIDTH),
    .ADDR_WIDTH         (ADDR_WIDTH),
    .AXIL_DATA_WIDTH    (AXIL_DATA_WIDTH)
) u_sdram_phy (
    .i_clk_sys          (i_clk_sys),
    .i_rst_sys          (i_rst_sys),

    .i_cmd              (phy_cmd),
    .i_ba               (phy_ba),
    .i_addr             (phy_addr),
    .i_data             (phy_data),
    .i_cmd_valid        (phy_cmd_valid),
    .o_cmd_ready        (phy_cmd_ready),

    .o_data             (phy_o_data),
    .o_data_valid       (phy_data_valid),
    .o_resp             (phy_resp),

    .o_sdram_dq         (o_sdram_dq),
    .o_sdram_dq_oe      (o_sdram_dq_oe),
    .i_sdram_dq         (i_sdram_dq),
    .o_sdram_addr       (o_sdram_addr),
    .o_sdram_ba         (o_sdram_ba),
    .o_sdram_cke        (o_sdram_cke),
    .o_sdram_cs_n       (o_sdram_cs_n),
    .o_sdram_ras_n      (o_sdram_ras_n),
    .o_sdram_cas_n      (o_sdram_cas_n),
    .o_sdram_we_n       (o_sdram_we_n)
);

endmodule