module sdram_phy #(
    parameter CAS_LATENCY=3,
    parameter BURST_LENGTH=2,
    parameter TRAS_NS=40,
    parameter TRC_NS=55,
    parameter TRCD_NS=15,
    parameter TRFC_NS=55,
    parameter TRP_NS=15,
    parameter TWR_NS=10,
    parameter TMRD_NS=10,

    parameter DQ_WIDTH=16,

    parameter BA_WIDTH=2,
    parameter ADDR_WIDTH=13,
    parameter AXIL_DATA_WIDTH = 32,

    localparam CMD_WIDTH=4
) (
    input  logic                        i_clk_sys,
    input  logic                        i_rst_sys,

    input  logic [CMD_WIDTH-1:0]        i_cmd,
    input  logic [BA_WIDTH-1:0]         i_ba,
    input  logic [ADDR_WIDTH-1:0]       i_addr,
    input  logic [AXIL_DATA_WIDTH-1:0]  i_data,
    input  logic                        i_cmd_valid,
    output logic                        o_cmd_ready,

    output logic [AXIL_DATA_WIDTH-1:0]  o_data,
    output logic                        o_data_valid,
    output logic [2:0]                  o_resp,


    output logic [DQ_WIDTH-1:0]         o_sdram_dq,
    output logic [DQ_WIDTH-1:0]         o_sdram_dq_oe,
    input  logic [DQ_WIDTH-1:0]         i_sdram_dq,
    output logic [ADDR_WIDTH-1:0]       o_sdram_addr,
    output logic [BA_WIDTH-1:0]         o_sdram_ba,
    output logic                        o_sdram_cke,
    output logic                        o_sdram_cs_n,
    output logic                        o_sdram_ras_n,
    output logic                        o_sdram_cas_n,
    output logic                        o_sdram_we_n
);

localparam CLK200_NS=5;

localparam TRAS_CLK=TRAS_NS/CLK200_NS;
localparam TRC_CLK=TRC_NS/CLK200_NS;
localparam TRCD_CLK=TRCD_NS/CLK200_NS;
localparam TRFC_CLK=TRFC_NS/CLK200_NS;
localparam TRP_CLK=TRP_NS/CLK200_NS;
localparam TWR_CLK=TWR_NS/CLK200_NS;
localparam TMRD_CLK=TMRD_NS/CLK200_NS;

logic [AXIL_DATA_WIDTH-1:0] data_reg, data_reg_next;

logic [31:0] counter, counter_next;
enum logic [1:0] {CMD, WAIT, READ, WRITE} state, state_next;
logic [3:0] cmd_reg, cmd_reg_next;

always_ff @(posedge i_clk_sys) begin
    if (i_rst_sys) begin
        counter <= '0;
        state <= CMD;
        cmd_reg <= '0;
        data_reg <= '0;
    end else begin
        counter <= counter_next;
        state <= state_next;
        cmd_reg <= cmd_reg_next;
        data_reg <= data_reg_next;
    end 
end

always_comb begin
    o_sdram_cs_n = '1;
    o_sdram_ras_n = '1;
    o_sdram_cas_n = '1;
    o_sdram_we_n = '1;

    o_sdram_cke = '1;

    o_sdram_ba = '0;
    o_sdram_addr = '0;
    o_sdram_dq = '0;
    o_sdram_dq_oe = '0;

    o_cmd_ready = '0;

    cmd_reg_next = cmd_reg;
    counter_next = counter;
    data_reg_next = data_reg;
    state_next = state;

    o_data = '0;
    o_data_valid = '0;
    o_resp = '0;

    case (state)
        CMD: begin
            if (i_cmd_valid) begin
                o_cmd_ready = '1;
                state_next = WAIT;

                o_sdram_cs_n = i_cmd[3];
                o_sdram_ras_n = i_cmd[2];
                o_sdram_cas_n = i_cmd[1];
                o_sdram_we_n = i_cmd[0];

                o_sdram_ba = i_ba;
                o_sdram_addr = i_addr;

                cmd_reg_next = i_cmd;

                case (i_cmd)
                    4'h0: begin
                        counter_next = TMRD_CLK;
                    end

                    4'h1: begin
                        counter_next = TRFC_CLK;
                    end

                    4'h2: begin
                        counter_next = TRP_CLK;
                    end

                    4'h3: begin
                        counter_next = TRCD_CLK;
                    end

                    4'h4: begin
                        state_next = WRITE;
                        o_sdram_dq = i_data[0 +: DQ_WIDTH];
                        o_sdram_dq_oe = '1;
                    end

                    4'h5: begin
                        counter_next = CAS_LATENCY;
                        state_next = WAIT;
                    end

                    default: begin
                        counter_next = '0;
                    end
                endcase
            end
        end

        WAIT: begin
            //nop, not deactivate
            o_sdram_cs_n = '0;

            // +2 because of transition clocks into and out of this state
            if (counter <= 2) begin
                if (cmd_reg == 4'h5) begin
                    state_next = READ;
                    counter_next = '0;
                end else begin
                    state_next = CMD;
                end
            end else begin
                counter_next = counter - 1;
            end
        end

        // This hard coded for 32 bit axi to 16 bit dram...
        WRITE: begin            
            o_sdram_cs_n = '0;

            o_sdram_dq = i_data[DQ_WIDTH +: DQ_WIDTH];
            o_sdram_dq_oe = '1;

            // the "burst length - 1" is handled in this state,
            // so we don't need to wait again
            counter_next = TWR_CLK+TRP_CLK;
            state_next = WAIT;

            o_data_valid = '1;
            o_resp = 3'b100;
        end

        READ: begin
            o_sdram_cs_n = '0;

            data_reg_next = {(AXIL_DATA_WIDTH-DQ_WIDTH)'(data_reg), i_sdram_dq};

            counter_next = counter + 1;

            if (counter == BURST_LENGTH-1) begin
                counter_next = TRP_CLK + BURST_LENGTH;   // I think this is pessimistic
                cmd_reg_next = 4'h8;  // nop
                state_next = WAIT;

                // can register this to help with timing
                o_data = data_reg_next;
                o_data_valid = '1;
                o_resp = '0;

                data_reg_next = '0;
            end
        end

        default: begin
            state_next = CMD;
        end
    endcase
end

endmodule